import { diffLines } from 'diff';
import { setCDN, getHighlighter } from 'shiki';
import React from 'react';
import { mapWithDefault } from '@code-hike/utils';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spread() {
    for (var ar = [], i = 0; i < arguments.length; i++)
        ar = ar.concat(__read(arguments[i]));
    return ar;
}

function tokenizePlaceholder(code) {
    var lines = code.split("\n");
    return lines.map(function (line) { return [[line, {}]]; });
}
function getBasicThemedCode(code, theme) {
    var lines = code.split("\n");
    var _a = getThemeDefaultColors(theme), fg = _a.fg, bg = _a.bg;
    return {
        lines: lines.map(function (line) { return [{ content: line }]; }),
        fg: fg,
        bg: bg,
    };
}
/**
 * From https://github.com/shikijs/shiki/blob/HEAD/packages/shiki/src/loader.ts
 */
var VSCODE_FALLBACK_EDITOR_FG = {
    light: "#333333",
    dark: "#bbbbbb",
};
var VSCODE_FALLBACK_EDITOR_BG = {
    light: "#fffffe",
    dark: "#1e1e1e",
};
function getThemeDefaultColors(theme) {
    var _a, _b, _c, _d, _e, _f;
    var fg, bg;
    /**
     * First try:
     * Theme might contain a global `tokenColor` without `name` or `scope`
     * Used as default value for foreground/background
     */
    var settings = theme.settings
        ? theme.settings
        : theme.tokenColors;
    var globalSetting = settings
        ? settings.find(function (s) {
            return !s.name && !s.scope;
        })
        : undefined;
    if ((_a = globalSetting === null || globalSetting === void 0 ? void 0 : globalSetting.settings) === null || _a === void 0 ? void 0 : _a.foreground) {
        fg = globalSetting.settings.foreground;
    }
    if ((_b = globalSetting === null || globalSetting === void 0 ? void 0 : globalSetting.settings) === null || _b === void 0 ? void 0 : _b.background) {
        bg = globalSetting.settings.background;
    }
    /**
     * Second try:
     * If there's no global `tokenColor` without `name` or `scope`
     * Use `editor.foreground` and `editor.background`
     */
    if (!fg && ((_d = (_c = theme) === null || _c === void 0 ? void 0 : _c.colors) === null || _d === void 0 ? void 0 : _d["editor.foreground"])) {
        fg = theme.colors["editor.foreground"];
    }
    if (!bg && ((_f = (_e = theme) === null || _e === void 0 ? void 0 : _e.colors) === null || _f === void 0 ? void 0 : _f["editor.background"])) {
        bg = theme.colors["editor.background"];
    }
    /**
     * Last try:
     * If there's no fg/bg color specified in theme, use default
     */
    if (!fg) {
        fg =
            theme.type === "light"
                ? VSCODE_FALLBACK_EDITOR_FG.light
                : VSCODE_FALLBACK_EDITOR_FG.dark;
    }
    if (!bg) {
        bg =
            theme.type === "light"
                ? VSCODE_FALLBACK_EDITOR_BG.light
                : VSCODE_FALLBACK_EDITOR_BG.dark;
    }
    return {
        fg: fg,
        bg: bg,
    };
}

function useAsyncMemo(_a, deps) {
    var isReady = _a.isReady, load = _a.load, run = _a.run, placeholder = _a.placeholder;
    var _b = __read(React.useState({ result: null, deps: [], version: 0 }), 2), _c = _b[0], lastLoadedResult = _c.result, lastLoadedDeps = _c.deps, version = _c.version, setState = _b[1];
    React.useEffect(function () {
        if (!isReady) {
            load().then(function () {
                setState(function (_a) {
                    var version = _a.version;
                    return ({
                        result: run(),
                        deps: deps,
                        version: version + 1,
                    });
                });
            });
        }
    }, deps);
    return React.useMemo(function () {
        if (depsChanged(deps, lastLoadedDeps)) {
            return isReady ? run() : placeholder();
        }
        else {
            return lastLoadedResult;
        }
    }, __spread([version], deps));
}
function depsChanged(oldDeps, newDeps) {
    if (oldDeps === newDeps)
        return false;
    if (!oldDeps || !newDeps)
        return true;
    return oldDeps === null || oldDeps === void 0 ? void 0 : oldDeps.some(function (dep, index) { return newDeps[index] !== dep; });
}

var _a;
var highlighterPromise = null;
var highlighter = null;
function useHighlighter(code, lang, theme) {
    return useAsyncMemo({
        isReady: isHighlighterReady(lang, theme),
        load: function () { return loadHighlighter(lang, theme); },
        run: function () { return highlight(code, lang, theme); },
        placeholder: function () {
            var _a = getBasicThemedCode("", theme), bg = _a.bg, fg = _a.fg;
            var lines = mapWithDefault(code, "", tokenizePlaceholder);
            return { lines: lines, bg: bg, fg: fg };
        },
    }, [code.prev, code.next, lang, theme]);
}
function isHighlighterReady(lang, theme) {
    return (highlighter != null &&
        !missingTheme(highlighter, theme) &&
        !missingLang(highlighter, lang));
}
function loadHighlighter(lang, theme) {
    return __awaiter(this, void 0, void 0, function () {
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (highlighterPromise === null) {
                        setCDN("https://unpkg.com/shiki/");
                        highlighterPromise = getHighlighter({
                            theme: theme,
                            langs: [lang],
                        });
                    }
                    if (!(highlighter === null)) return [3 /*break*/, 2];
                    return [4 /*yield*/, highlighterPromise];
                case 1:
                    highlighter = _a.sent();
                    _a.label = 2;
                case 2:
                    if (!missingTheme(highlighter, theme)) return [3 /*break*/, 4];
                    return [4 /*yield*/, highlighter.loadTheme(theme)];
                case 3:
                    _a.sent();
                    _a.label = 4;
                case 4:
                    if (!missingLang(highlighter, lang)) return [3 /*break*/, 6];
                    return [4 /*yield*/, highlighter.loadLanguage(lang)];
                case 5:
                    _a.sent();
                    _a.label = 6;
                case 6: return [2 /*return*/];
            }
        });
    });
}
function highlight(code, lang, theme) {
    // assumes highlighter isReady
    var _a = highlighter.getTheme(theme.name), fg = _a.fg, bg = _a.bg;
    var lines = mapWithDefault(code, "", function (code) {
        return getCodeLines(highlighter, code, lang, theme);
    });
    return { lines: lines, bg: bg, fg: fg };
}
function getCodeLines(highlighter, code, lang, theme) {
    var lines = highlighter.codeToThemedTokens(code, lang, theme.name, {
        includeExplanation: false,
    });
    return lines.map(function (line) {
        return line.map(function (token) { return [
            token.content,
            { style: getStyle(token) },
        ]; });
    });
}
function missingTheme(highlighter, theme) {
    return !highlighter
        .getLoadedThemes()
        .some(function (t) { return t === theme.name; });
}
function missingLang(highlighter, lang) {
    return !highlighter
        .getLoadedLanguages()
        .some(function (l) { return l === lang; });
}
var FONT_STYLE_TO_CSS = (_a = {},
    _a[-1 /* NotSet */] = {},
    _a[0 /* None */] = {},
    _a[1 /* Italic */] = { fontStyle: "italic" },
    _a[2 /* Bold */] = { fontWeight: "bold" },
    _a[4 /* Underline */] = { textDecoration: "underline" },
    _a);
function getStyle(token) {
    var fontStyle = token.fontStyle
        ? FONT_STYLE_TO_CSS[token.fontStyle]
        : {};
    return __assign({ color: token.color }, fontStyle);
}

function useCodeDiff(_a) {
    var code = _a.code, lang = _a.lang, theme = _a.theme;
    var normalCode = mapWithDefault(code, "", normalize);
    var _b = useHighlighter(normalCode, lang, theme), lines = _b.lines, bg = _b.bg, fg = _b.fg;
    return React.useMemo(function () {
        var prevKeys = Array.from({ length: lines.prev.length - 1 }, function (x, i) { return i + 1; });
        var nextKeys = getNextKeys(normalCode.prev, normalCode.next, prevKeys);
        var codeMap = {
            keys: { prev: prevKeys, next: nextKeys },
            lines: {},
        };
        prevKeys.forEach(function (key, index) {
            return (codeMap.lines[key] = {
                tokens: lines.prev[index],
                lineNumber: { prev: index + 1 },
            });
        });
        nextKeys.forEach(function (key, index) {
            if (key in codeMap.lines) {
                codeMap.lines[key].lineNumber.next = index + 1;
            }
            else {
                codeMap.lines[key] = {
                    tokens: lines.next[index],
                    lineNumber: { next: index + 1 },
                };
            }
        });
        return {
            codeMap: codeMap,
            backgroundColor: bg,
            color: fg,
        };
    }, [lines.prev, lines.next, bg, fg]);
}
function getNextKeys(prevCode, nextCode, prevKeys) {
    var changes = diffLines(prevCode, nextCode);
    var nextKeys = [];
    var prevIndex = 0;
    changes.forEach(function (change) {
        if (change.added) {
            var p = prevKeys[prevIndex - 1] != null
                ? prevKeys[prevIndex - 1]
                : prevKeys[prevIndex] - 1;
            var n = prevKeys[prevIndex] != null
                ? prevKeys[prevIndex]
                : prevKeys[prevKeys.length - 1] + 1;
            for (var i = 1; i <= change.count; i++) {
                nextKeys.push(p + ((n - p) * i) / (change.count + 1));
            }
        }
        else if (change.removed) {
            prevIndex += change.count;
        }
        else {
            for (var i = 1; i <= change.count; i++) {
                nextKeys.push(prevKeys[prevIndex]);
                prevIndex++;
            }
        }
    });
    return nextKeys;
}
function normalize(text) {
    return (text || "").trimEnd().concat("\n");
}

export { useCodeDiff };
